import unittest
import litemapy.storage as storage
import math
from . import helper

@unittest.skipUnless(helper.java_test_available(), "Cannot run java in this environment")
class TestAgainstJavaLitematica(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.subproc, cls.gateway = helper.get_litematica_jvm()
        cls.JLitematicaBitArray = cls.gateway.jvm.fi.dy.masa.litematica.schematic.container.LitematicaBitArray

    @classmethod
    def tearDownClass(cls):
        helper.terminate_litematica_jvm()

    def setUp(self):
        pass

    def test_import_from_litematica(self):
        nbits = 6
        size = 33
        jarr = self.JLitematicaBitArray(nbits, size)
        for i in range(size):
            jarr.setAt(i, i)
        longarr = jarr.getBackingLongArray()
        parr = storage.LitematicaBitArray.fromnbtlongarray(longarr, size, nbits)
        for i in range(size):
            self.assertEqual(parr[i], i)

    def test_export_to_litematica(self):
        nbits = 6
        size = 33
        parr = storage.LitematicaBitArray(size, nbits)
        for i in range(size):
            parr[i] = i
        plongs = parr._tolonglist()
        jlongs = self.gateway.new_array(self.gateway.jvm.long, len(plongs))
        for i, l in enumerate(plongs):
            jlongs[i] = l
        jarr = self.JLitematicaBitArray(nbits, size, jlongs)
        for i in range(len(parr)):
            self.assertEqual(parr[i], jarr.getAt(i))



import unittest
import litemapy.boxes as boxes

class TestBoxes(unittest.TestCase):

    def setUp(self):
        self.origin = (0, 0, 0)
        self.block1 = (-2, -2, -2)
        self.block2 = (2, 2, 2)
        self.block3 = (1, 2, 3)
        self.block4 = (-4, -4, -4)
        self.block5 = (4, 4, 4)
        self.centeredbox1 = (self.block1, self.block2)
        self.centeredbox2 = (self.block4, self.block5)
        self.box1 = (self.origin, self.block2)
        self.box2 = (self.block1, self.block5)
        self.box3 = (self.block1, self.block4)

    def test_block_is_in_box(self):
        self.assertTrue(boxes.block_is_in_box(self.origin, self.centeredbox1))
        self.assertFalse(boxes.block_is_in_box(self.block3, (self.block1, self.block2)))

    def test_box_is_in_box(self):
        self.assertTrue(boxes.box_is_in_box(self.centeredbox1, self.centeredbox2))
        self.assertTrue(boxes.box_is_in_box(self.box1, self.centeredbox2))
        self.assertFalse(boxes.box_is_in_box(self.centeredbox2, self.centeredbox1))
        self.assertFalse(boxes.box_is_in_box(self.box1, self.box3))



LITEMATIC_VERSION = 5 # Litematic version we are trying to imitate from Litematica
MC_DATA_VERSION = 1631
DEFAULT_NAME = "Unnamed" # Default name given to schematics and regions if unspecified
LITEMAPY_NAME = "Litemapy" # Used to identify schematic created with Litemapy in metadata
LITEMAPY_VERSION = "0.2.1a0" # Unique version string



from math import ceil
import nbtlib.tag

class LitematicaBitArray:

    def __init__(self, size, nbits):
        self.size = size
        self.nbits = nbits
        s = ceil(nbits * size / 64)
        self.array = [0] * s
        self.__mask = (1 << nbits) - 1 # nbits bits set to 1

    def fromnbtlongarray(arr, size, nbits):
        excpected_len = ceil(size * nbits / 64)
        if excpected_len != len(arr):
            raise ValueError(
                "long array length does not match bit array size and nbits, excpected {}, not {}".format(
                        excpected_len, len(arr)
                    )
                )
        r = LitematicaBitArray(size, nbits)
        m = (1 << 64) - 1
        r.array = [int(i) & m for i in arr] # Remove the infinite trailing 1s of negative numbers
        return r

    def _tolonglist(self):
        l = []
        m1 = 1 << 63
        m2 = (1 << 64) - 1
        for i in self.array:
            if i & m1 > 0:
                i |= ~m2 # Add the potential infinit 1 prefix for negative numbers
            l.append(i)
        return l

    def _tonbtlongarray(self):
        return nbtlib.tag.LongArray(self._tolonglist())

    def __getitem__(self, index):
        if not 0 <= index < len(self):
            raise IndexError("Invalid index {}".format(index))
        startOffset = index * self.nbits
        startArrIndex = startOffset >> 6
        endArrIndex = ((index + 1) * self.nbits - 1) >> 6
        startBitOffset = startOffset & 0x3F

        if startArrIndex == endArrIndex : 
            return self.array[startArrIndex] >> startBitOffset & self.__mask
        else:
            endOffset = 64 - startBitOffset;
            return (self.array[startArrIndex] >> startBitOffset | self.array[endArrIndex] << endOffset) & self.__mask

    def __setitem__(self, index, value):
        if not 0 <= index < len(self):
            raise IndexError("Invalid index {}".format(index))
        if not 0 <= value <= self.__mask:
            raise ValueError("Invalid value {}, maximum value is {}".format(value, self.__mask))
        startOffset = index * self.nbits
        startArrIndex = startOffset >> 6
        endArrIndex = ((index + 1) * self.nbits - 1) >> 6
        startBitOffset = startOffset & 0x3F
        m = (1 << 64) - 1
        self.array[startArrIndex] = (self.array[startArrIndex] & ~(self.__mask << startBitOffset) | (value & self.__mask) << startBitOffset) & m

        if startArrIndex != endArrIndex:
            endOffset = 64 - startBitOffset;
            j1 = self.nbits - endOffset;
            self.array[endArrIndex] = (self.array[endArrIndex] >> j1 << j1 | ( value & self.__mask) >> endOffset) & m

    def __len__(self):
        return self.size

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __reversed__(self):
        arr = LitematicaBitArray(self.size, self.nbits)
        for i in range(len(self)):
            arr[i] = self[len(self) - i - 1]
        return arr
    
    def __contains__(self, value):
        for v in self:
            if v == value:
                return True
        return False

class DiscriminatingDictionnary(dict):

    def __init__(self, validator, *args, **options):
        """
        validator: a function that takes as argument a key and an item and returns a tuple (canstore, msg)
        canstore must be a boolean, True if the item is accepted, and False otherwwise
        if canstore is false, msg will be used as the error message
        """
        #TODO Handle iterators in constructor
        self.validator = validator
        if len(args) == 1 and type(args[0]) == dict:
            for key, item in args[0].items():
                self.validate(key, item)
        else:
            for key, item in args:
                self.validate(key, item)
        for key, item in options.items():
            self.validate(key, item)
        super().__init__(*args, **options)

    def validate(self, key, item):
        canstore, msg = self.validator(key, item)
        if not canstore:
            raise DiscriminationError(msg)

    def __setitem__(self, key, item):
        self.validate(key, item)
        super().__setitem__(key, item)

    def setdefault(key, default=None):
        self.validate(key, default)
        super().setdeault(key, default)

    def update(self, other):
        other = DiscriminatingDictionnary(self.validator, other)
        super().update(other)

class DiscriminationError(Exception):
    pass



import os
import subprocess
import py4j.java_gateway
from .constants import *

SUB_PROC, GATEWAY = None, None

def java_test_available():
    return not is_windows() #TODO

def get_litematica_jvm():
    """
    The subprocess needs to be terminated!
    """
    global SUB_PROC, GATEWAY
    if SUB_PROC == None:
        gwrapper_path = JAVA_TEST_PROJECT + "/gradlew"
        os.chmod(gwrapper_path, 0b0111110100) #May need to be changed
        cmd = [gwrapper_path, "-p", JAVA_TEST_PROJECT, "--console=plain", "run"]
        SUB_PROC = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        line = SUB_PROC.stdout.readline().decode('utf-8')
        while line is not None:
            if "[JAVA] Gateway Server Started..." in line:
                break
            line = SUB_PROC.stdout.readline().decode('utf-8')
        GATEWAY = py4j.java_gateway.JavaGateway()
    return SUB_PROC, GATEWAY

def terminate_litematica_jvm():
    print("Closing gateway")
    GATEWAY.close()
    try:
        SUB_PROC.wait(1)
    except:
        pass # This is excepted, we are only letting some time for the gateway to start
    print("Killing java subprocess")
    SUB_PROC.kill()
    try:
        SUB_PROC.wait(1)
    except:
        pass # This is excepted, we are only letting some time for the gateway to start

def is_windows():
    return os.name == 'nt'



def block_is_in_box(block, box):
    """
    Returns True id the block is in the box
    block is (x, y, z)
    box is ((x0, y0, z0), (x1, y1, z1))
    """
    x, y, z = block
    xs = [box[0][0], box[1][0]]
    xs.sort()
    ys = [box[0][1], box[1][1]]
    ys.sort()
    zs = [box[0][1], box[1][1]]
    zs.sort()
    xmin, xmax = xs
    ymin, ymax = ys
    zmin, zmax = zs
    return xmin <= x <= xmax and ymin <= y <= ymax and zmin <= z <= zmax

def box_is_in_box(box1, box2):
    return block_is_in_box(box1[0], box2) and block_is_in_box(box1[1], box2)



from nbtlib.tag import End, Byte, Short, Int, Long, Float, Double, ByteArray, String, List, Compound, IntArray, LongArray
import nbtlib
from math import ceil, log
from .storage import LitematicaBitArray, DiscriminatingDictionnary
from .info import *
from .boxes import *
from time import time

class Schematic:

    """
    A schematic file
    """

    def __init__(self,
                    width, height, length, 
                    name=DEFAULT_NAME, author="", description="",
                    regions={}, main_region_name=DEFAULT_NAME
                ):
        """
        Initialize a schematic of size width, height and length
        name, author and description are used in metadatas
        regions should be disctionnary {'regionname': region} to add to the schematic
        If region is None or empty, an empty region, named after main_region_name,
        and of the size of the schematic is created at its origin
        """
        self.author = author
        self.description = description
        self.name = name
        self.created = int(time())
        self.modified = int(time())
        self.__width, self.__height, self.__length = width, height, length
        self.__regions = DiscriminatingDictionnary(self._can_add_region)
        if regions != None and len(regions) > 0:
            self.__regions.update(regions)
        else:
            reg = Region(0, 0, 0, self.width, self.height, self.length)
            self.__regions[main_region_name] = reg

    def save(self, fname, update_meta=True, save_soft=True):
        """
        Save this schematic to the disk in a file name fname
        update_meta: update metadata before writing to the disk (modified time)
        save_soft: add a metadata entry with the software name and version
        """
        if update_meta:
            self.updatemeta()
        f = nbtlib.File(gzipped=True, byteorder='big')
        f[""] = self._tonbt(save_soft=save_soft)
        f.save(fname)

    def _tonbt(self, save_soft=True):
        """
        Write the schematic to an nbt tag
        """
        root = Compound()
        root["Version"] = Int(LITEMATIC_VERSION)
        root["MinecraftDataVersion"] = Int(MC_DATA_VERSION)
        meta = Compound()
        enclose = Compound()
        enclose["x"] = Int(self.width)
        enclose["y"] = Int(self.height)
        enclose["z"] = Int(self.length)
        meta["EnclosingSize"] = enclose
        meta["Author"] = String(self.author)
        meta["Description"] = String(self.description)
        meta["Name"] = String(self.name)
        meta["Software"] = String(LITEMAPY_NAME + "_" + LITEMAPY_VERSION)
        meta["RegionCount"] = Int(len(self.regions))
        meta["TimeCreated"] = Long(self.created)
        meta["TimeModified"] = Long(self.modified)
        meta["TotalBlocks"] = Int(sum([reg.getblockcount() for reg in self.regions.values()])) 
        meta["TotalVolume"] = Int(sum([reg.getvolume() for reg in self.regions.values()]))
        root["Metadata"] = meta
        regs = Compound()
        for regname, reg in self.regions.items():
            regs[regname] = reg._tonbt()
        root["Regions"] = regs
        return root

    def fromnbt(nbt):
        """
        Read and return a schematic from an nbt tag
        """
        meta = nbt["Metadata"]
        width = int(meta["EnclosingSize"]["x"])
        height = int(meta["EnclosingSize"]["y"])
        length = int(meta["EnclosingSize"]["z"])
        author = nbtstr2str(meta["Author"])
        name = nbtstr2str(meta["Name"])
        desc = nbtstr2str(meta["Description"])
        regions = {}
        for key, value in nbt["Regions"].items():
            reg = Region.fromnbt(value)
            regions[str(key)] = reg
        sch = Schematic(width, height, length, name=name, author=author, description=desc, regions=regions)
        sch.created = int(meta["TimeCreated"])
        sch.modified = int(meta["TimeModified"])
        if "RegionCount" in meta and len(sch.regions) != meta["RegionCount"]:
            raise CorruptedSchematicError("Number of regions in metadata does not match the number of parsed regions")
        return sch

    def updatemeta(self):
        """
        Update this schematic's metadata (modified time)
        """
        self.modified = int(time())

    def load(fname):
        """
        Read a schematic from disk
        fname: name of the file
        """
        nbt = nbtlib.File.load(fname, True)['']
        return Schematic.fromnbt(nbt)

    def _can_add_region(self, name, region):
        if type(name) != str:
            return False, "Region name should be a string"
        schembox = ((0, 0, 0), (self.width-1, self.height-1, self.length-1))
        regx, regy, regz = region.x, region.y, region.z
        regbx = regx + region.width - abs(region.width) / region.width
        regby = regy + region.height - abs(region.height) / region.height
        regbz = regz + region.length - abs(region.length) / region.length
        regbox = ((regx, regy, regz), (regbx, regy, regz))
        if not box_is_in_box(regbox, schembox):
            return False, "The region does not fit in the schematic"
        return True, ""

    @property
    def regions(self):
        return self.__regions
    
    @property
    def width(self):
        return self.__width

    @property
    def height(self):
        return self.__height

    @property
    def length(self):
        return self.__length

class Region:

    """
    A schematic region
    x, y, z: position in the schematic (read only)
    width, height, length: size of the region (oriented, can be negative)
    """

    def __init__(self, x, y, z, width, height, length):
        self.__x, self.__y, self.__z = x, y, z
        self.__width, self.__height, self.__length = width, height, length
        self.__palette = [AIR, ]
        self.__blocks = [[[ 0 for k in range(abs(length))] for j in range(abs(height))] for i in range(abs(width))]
        self.entities = [] #TODO Add support
        self.tileentities = [] #TODO Add support

    def _tonbt(self):
        """
        Write this region to an nbt tab and return it
        """
        root = Compound()
        pos = Compound()
        pos["x"] = Int(self.__x)
        pos["y"] = Int(self.__y)
        pos["z"] = Int(self.__z)
        root["Position"] = pos
        size = Compound()
        size["x"] = Int(self.__width)
        size["y"] = Int(self.__height)
        size["z"] = Int(self.__length)
        root["Size"] = size
        plt = List[Compound]([blk._tonbt() for blk in self.__palette])
        root["BlockStatePalette"] = plt
        ents = List[Compound]([ent._tonbt() for ent in self.entities])
        root["Entities"] = ents
        tilents = List[Compound]([ent._tonbt for ent in self.tileentities])
        root["TileEntities"] = tilents
        root["PendingBlockTicks"] = List[Compound]() #TODO How does this work
        root["PendingFluidTicks"] = List[Compound]()
        arr = LitematicaBitArray(self.getvolume(), self.__get_needed_nbits())
        for x in range(abs(self.__width)):
            for y in range(abs(self.__height)):
                for z in range(abs(self.__length)):
                    ind = (y * abs(self.__width * self.__length)) + z * abs(self.__width) + x
                    arr[ind] = self.__blocks[x][y][z]
        root["BlockStates"] = arr._tonbtlongarray()
        return root

    def getblock(self, x, y, z):
        """
        Return the block at the given coordinates
        """
        if self.__width < 0:
            x -= self.__width + 1
        if self.__height < 0:
            y -= self.__height + 1
        if self.__length < 0:
            z -= __self.length + 1
        return self.__palette[ self.__blocks[x][y][z] ]

    def setblock(self, x, y, z, block):
        """
        Set the block at the given coordinate
        """
        if block in self.__palette:
            i = self.__palette.index(block)
        else:
            self.__palette.append(block)
            i = len(self.__palette) - 1
        self.__blocks[x][y][z] = i

    def getblockcount(self):
        """
        Returns the number of non-air in the region
        """
        airind = self.__palette.index(AIR)
        c = 0
        for plan in self.__blocks:
            for column in plan:
                for block in column:
                    if block != airind:
                        c += 1
        return c

    def getvolume(self):
        """
        Returns the region's volume
        """
        return abs(self.__width * self.__height * self.__length)

    def __get_needed_nbits(self):
        return max(ceil(log(len(self.__palette), 2)), 2)
    
    def fromnbt(nbt):
        """
        Read a region from an nbt tag and return it
        """
        pos = nbt["Position"]
        x = int(pos["x"])
        y = int(pos["y"])
        z = int(pos["z"])
        size = nbt["Size"]
        width = int(size["x"])
        height = int(size["y"])
        length = int(size["z"])
        reg = Region(x, y, z, width, height, length)
        del reg.__palette[0]
        for bnbt in nbt["BlockStatePalette"]:
            block = BlockState.fromnbt(bnbt)
            reg.__palette.append(block)
        for enbt in nbt["Entities"]:
            entity = Entity.fromnbt(entity)
            reg.entities.append(entity)
        for tenbt in nbt["TileEntities"]:
            block = TileEntity.fromnbt(tentity)
            reg.tileentities.append(block)
        blks = nbt["BlockStates"]
        nbits = reg.__get_needed_nbits()
        arr = LitematicaBitArray.fromnbtlongarray(blks, reg.getvolume(), nbits)
        for x in range(abs(width)):
            for y in range(abs(height)):
                for z in range(abs(length)):
                    ind = (y * abs(width * length)) + z * abs(width) + x
                    reg.__blocks[x][y][z] = arr[ind]
        return reg

    @property
    def x(self):
        return self.__x

    @property
    def y(self):
        return self.__y

    @property
    def z(self):
        return self.__z

    @property
    def width(self):
        return self.__width

    @property
    def height(self):
        return self.__height

    @property
    def length(self):
        return self.__length



class BlockState:

    def __init__(self, blockid, properties={}):
        self.blockid = blockid
        self.properties = {String(k): String(v) for k, v in properties.items()}

    def _tonbt(self):
        root = Compound()
        root["Name"] = String(self.blockid)
        if len(self.properties) > 0:
            root["Properties"] = Compound(self.properties)
        return root

    def fromnbt(nbt):
        block = BlockState(nbtstr2str(nbt["Name"]))
        for key, value in nbt.items():
            block.properties[nbtstr2str(key)] = nbtstr2str(value)
        return block

class Entity:

    def __init__(self):
        pass #TODO

    def _tonbt(self):
        raise NotImplementedError("Entities are not supported yet")

class TileEntity:

    def __init__(self):
        pass #TODO

    def _tonbt(self):
        raise NotImplementedError("Tile entities are not supported yet")

def nbtstr2str(s):
    return str(s)[1:-1]

AIR = BlockState("minecraft:air")

class CorruptedSchematicError(Exception):
    pass




import unittest
import litemapy.storage as storage
import math
from . import helper

class TestLitematicaBitArray(unittest.TestCase):

    def test_write_read(self):
        l = [0, 0, 0, 12, 13, 0, 4, 0, 2, 4, 1, 3, 3, 7, 65, 9]
        nbits = math.ceil(math.log(max(l), 2)) + 1
        arr = storage.LitematicaBitArray(len(l), nbits)
        for i, e in enumerate(l):
            arr[i] = e
        for i, e in enumerate(l):
            self.assertEqual(e, arr[i])

    def test_exceptions(self):
        arr = storage.LitematicaBitArray(10, 4)
        def setat(i):
            arr[i] = 0
        def setval(v):
            arr[0] = v
        self.assertRaises(IndexError, setat, -1)
        self.assertRaises(IndexError, setat, 10)
        self.assertRaises(ValueError, setval, -1)
        self.assertRaises(ValueError, setval, 16)

    def test_in(self):
        l = [0, 0, 0, 12, 13, 0, 4, 0, 2, 4, 1, 3, 3, 7, 65, 9]
        nbits = math.ceil(math.log(max(l), 2)) + 1
        arr = storage.LitematicaBitArray(len(l), nbits)
        for i, e in enumerate(l):
            arr[i] = e
        self.assertIn(13, arr)
        self.assertNotIn(15, arr)



#!/usr/bin/env python3

import setuptools
from litemapy import __version__ as LITEMAPY_VERSION

DESCRIPTION = "Read and write Litematica's Minecraft schematics files"

def readme():
    with open("README.md") as f:
        txt = f.read()
    return txt


setuptools.setup(
        name="litemapy",
        version=LITEMAPY_VERSION,
        author="SmylerMC",
        author_email="smyler@mail.com",
        description=DESCRIPTION,
        long_description=readme(),
        long_description_content_type="text/markdown",
        url="https://github.com/SmylerMC/litemapy",
        packages=setuptools.find_packages(exclude=["tests"]),
        license="GNU General Public License v3 (GPLv3)",
        classifiers=[
                "Development Status :: 3 - Alpha",
                "Intended Audience :: Other Audience",
                "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
                "Operating System :: OS Independent",
                "Programming Language :: Python :: 3",
                "Topic :: Games/Entertainment",
            ],
        python_requires=">=3.5",
        install_requires=[
                'nbtlib',
          ],
        test_suite='nose.collector',
        tests_require=['nose'],
    )
